# class TM for real-time TM, entropy

import sys, time
import math
import numpy as np
import scipy
from numpy.linalg import eig
from scipy.sparse.linalg import eigs
from collections import deque

# prevent Python from writing .pyc files
sys.dont_write_bytecode = True

from point import Point
from fixation import Fixation
from bw import Butterworth as BW 
from sg import SG
from monitor import Monitor
from grid import Grid
from statistics import mean, stdev, median

class TM:
  
    def __init__(self,cols,rows):

        # how many AOI (states) (cols x rows), use integers [0,n-1]
        self.states = list(range(cols * rows))
        print("states: ", self.states)

        self.grid = Grid(cols,rows)

        # this "memory" will have a large effect on how quickly TM gets
        # updated (cleared out)
        # let's assume rhythmic theory of attention: 2-8 Hz, meaning
        # 125-500 ms sampling period, maybe set to 125?
#       self.win = 500
#       self.win = 250
        # this is ok if we update seq every transition
        self.win = 125
#       self.win = 66
#       self.win = 33
#       self.win = 30
        # this is ok if we update seq upon sac->fxtn transitions
#       self.win = 10
        # initialize deque with all -1
#       self.seq = deque([-1] * self.win,maxlen=self.win)
        # initialize empty dequeu
        self.seq = deque(maxlen=self.win)

        self.h_t = 0.0		        # current transition entropy
        self.h_s = 0.0		        # current stationary transition entropy

        self.bwf_x = BW(2,30,1.15)  # Butterworth filter init
        self.bwf_y = BW(2,30,1.15)  # Butterworth filter init

        self.sgf_x = SG(5,2,1)      # Savitzky-Golay init (width,degree,op)
        self.sgf_y = SG(5,2,1)      # Savitzky-Golay init (width,degree,op)

        self.state = 0              # 0: saccade, 1: fixation
        self.threshold = 30.0       # velocity threshold (30 deg/s)

        # display arguments: x,y,w,h,dist,n, window size we want
        self.monitor = Monitor(0,0,1680,1050,22,19.68)

        # initialize transition matrix
        self.tmat = np.zeros((len(self.states),len(self.states)))
        self.nmat = np.zeros((len(self.states),len(self.states)))

    def update(self):

        # reset transition matrix
        self.tmat = np.zeros((len(self.states),len(self.states)))
        self.nmat = np.zeros((len(self.states),len(self.states)))

        # shorthand for number of rows
        siz = len(self.tmat)

        # examine sequence
#       print("seq: ", self.seq)

        # compute entropy
        if len(self.seq) > 0:

            # update TM
            for i in range(len(self.seq)-1):
                i1 = self.seq[i]
                i2 = self.seq[i+1]
                # omit transition to and from bogus AOIs (-1)
                if i1 < 0 or i2 < 0:
                    pass
                # omit self-transitions
#               elif i1 == i2:
#                   pass
                else:
#                   print("i1, i2: ", i1, i2)
                    self.tmat[i1,i2] += 1

            # normalize transition probabilities (per row)
            for i in range(siz):
                srow = np.sum(self.tmat[i,:])
#               print("srow: ", srow)
                if abs(srow) < 0.00001:
                    # CONSTRUCTION ASSUMPTION
                    # if no observations, set to 1/n to maximize row entropy
                    # counter-intuitive for real-time implementation
#                   self.nmat[i,:] = 1.0/siz
                    # if no observations, set to 0 so row doesn't get counted
                    self.nmat[i,:] = 0.0
                else:
                    for j in range(siz):
                        self.nmat[i,j] = self.tmat[i,j]/srow
                        if math.isnan(self.nmat[i,j]):
                            self.nmat[i,j] = 0.0
#           print("tmat with for-loop normalization:\n", self.nmat)

        else:
            # sequence empty
            print("seq empty")

            # normalize transition probabilities (per row) to max entropy
            for i in range(siz):
                # every row should sum up to zero
                srow = np.sum(self.tmat[i,:])
#               print("srow: ", srow)
                if abs(srow) < 0.00001:
                    # if no observations, set to max entropy
                    self.nmat[i,:] = 1.0/siz

    def H_t(self):

        if len(self.nmat) > 0:

            siz = len(self.nmat)
            max_entrop = np.log2(siz)
#           print("siz: ", siz)
#           print("len(self.nmat): ", len(self.nmat))
            print("starting max_entrop: ", max_entrop)

            H = np.zeros(siz)
#           print("H: ",H)

            # counter to see how many non-zero rows we have
            s = 0
            for i in range(siz):
                srow = np.sum(self.nmat[i,:])
#               print("srow: ", srow)
                if abs(srow) < 0.00001:
#                   print("got a zero row")
                    pass
                else:
                    # accumulate number of non-zero rows
                    s += 1
                    # compute entropy for this non-zero row
#                   print("i, H: ",i, H)
                    for j in range(siz):
                        if abs(self.nmat[i,j]) > 0.0001:
                            H[i] += self.nmat[i,j] * np.log2(self.nmat[i,j])
#                           print("H[%d]: %f" % (i,-H[i]))

            if s > 1:
                max_entrop = np.log2(s)
            print("ending max_entrop: ", max_entrop)
            print("H: %f, H normalized: %f" % (-np.sum(H),-np.sum(H)/siz))
#           self.h_t = abs(-np.sum(H)/siz)
            self.h_t = abs(-np.sum(H)/(siz * max_entrop))
            print("H_t: ",self.h_t)
        else:
            self.h_t = -1

        return self.h_t

    def H_s(self):

        if len(self.nmat) > 0:

            siz = len(self.nmat)
            max_entrop = np.log2(siz)
#           print("siz: ", siz)
#           print("len(self.nmat): ", len(self.nmat))
#           print("max_entrop: ", max_entrop)


            ''' using scipy sparse linalg eigs
            '''
            # alternate way of doing it, more direct
            lmx, vmx = eigs(self.nmat, k=1, which='LM')
            # reshape and extract vector from [[ ... ]]
            vmx = vmx.real.reshape(1,siz)[0]
#           print("vmx: ", vmx)

            # this gives us \pi, stationary distribution from the eigenvector
#           mag = np.linalg.norm(vmx)
#           print("mag(vmx): ", mag)
            smx = np.sum(vmx)
#           print("sum(vmx): ", smx)
            pmx = vmx * 1.0/smx
#           print("pmx: ", pmx)

            # $\sum_{i}{\pi_i}$ should equal 1.0
#           print("np.sum(pmx) [should be equal to 1.0]: ", np.sum(pmx))

            if abs(np.sum(pmx) - 1.0) > 0.0001:
            	print("Warning: np.sum(pmx) is not equal to 1.0")

            # see: http://www-isl.stanford.edu/~cover/papers/paper101.pdf
            # really should do H(M) = -\sum_i \pi_i \sum_j M_{ij} log_2(M_{ij})

            # using pmx
            H = 0.0
#           print("H: ",H)
            for i in range(siz):
              if abs(pmx[i]) > 0.0001:
#                 H += (pmx[i] * np.log2(pmx[i]))
                  H += (pmx[i] * np.log2(abs(pmx[i])))
#             print("%d: H: %f" % (i,-H))

            H = -H/max_entrop
            self.h_s = abs(H)
            print("H_s: ",H)

        else:
            self.h_s = -1

        return self.h_s

    def smooth(self,x,y,t):
        # smooth points
        bwx = self.bwf_x.bwlpf(x)
        bwy = self.bwf_y.bwlpf(y)

        return (bwx, bwy)

    def filter(self,x,y,t,dt):
        # t: current timestamp, dt: time period between samples (ms)

        # velocity in screen coords (normalized)
        dx, dy = self.sgf_x.sgf(x), self.sgf_y.sgf(y)

        # sampling period in sec
        period = float(dt)/1000.0
        
        # multiply by sgf length
        dt = period * float(self.sgf_x.n)

        # use monitor to compute degrees, note monitor uses screen dimensions
        deg_x, deg_y = self.monitor.pix2deg(dx), self.monitor.pix2deg(dy)

        # deg_x, deg_y are in deg per filter window, div by dt for per sec
        vel_x = deg_x / dt
        vel_y = deg_y / dt

        vel = math.sqrt(vel_x*vel_x + vel_y*vel_y)

        # print("velocity (pixels) = (%f,%f)" % (dx,dy))
        # print("velocity (deg) = (%f,%f)" % (velx,vely))
        # print("velocity (deg) = (%f)" % (vel))

        return vel

    def classify(self,AOI,x,y,t,dt,vel):

        if AOI < 0:
            print("classify: AOI: %d, removing from sequence" % (AOI))
            # outside any legit AOIs, remove from seq
            if len(self.seq):
                self.seq.popleft()
            return

        # state table
        # state         |   input    |  output
        # ------------------------------------
        # (1) fixation  | a < T (1)  |    (1) fixation
        # (0) saccade   | a < T (1)  |    (1) fixation
        # (1) fixation  | a > T (0)  |    (0) saccade
        # (0) saccade   | a > T (0)  |    (0) saccade
        #

        # amplitude > T (saccade)
        if math.fabs(vel) > self.threshold:
            if self.state == 1:
                # state transition frmo fixation to saccade
                self.fxtn_sacc(AOI,x,y,t,dt)
            else:
                # state transition frmo saccade to saccade
                self.sacc_sacc(AOI,x,y,t,dt)
            self.state = 0
        # amplitude < T (fixation)
        else:
            if self.state == 1:
                # state transition frmo fixation to fixation
                self.fxtn_fxtn(AOI,x,y,t,dt)
            else:
                # state transition frmo saccade to fixation
                self.sacc_fxtn(AOI,x,y,t,dt)
            self.state = 1

        # big question is when to add to the sequence of states
        # here, we can do this every gaze point
#       self.seq.append(AOI)

        return self.state

    def sacc_sacc(self,AOI,x,y,t,dt):
        # transition from saccade to saccade
#       self.seq.append(AOI)

        pass

    def fxtn_fxtn(self,AOI,x,y,t,dt):
        # transition from fixation to fixation
        self.seq.append(AOI)

        pass

    def sacc_fxtn(self,AOI,x,y,t,dt):
        # transition from saccade to fixation

        # here, we only add at saccaded -> fixation transition
        self.seq.append(AOI)
        pass

    def fxtn_sacc(self,AOI,x,y,t,dt):
        # transition from fixation to saccade

        self.seq.append(AOI)
        pass
